from pathlib import Path
from typing import Union, Optional, List
import numpy as np
import dask.array as da
import zarr
from wsireg.utils.im_utils import ARRAYLIKE_CLASSES, TIFFFILE_EXTS
from wsireg.parameter_maps.preprocessing import ImagePreproParams
from . import CziRegImage  # AICSRegImage,
from . import NumpyRegImage, SitkRegImage, TiffFileRegImage


def reg_image_loader(
    image: Union[np.ndarray, da.Array, zarr.Array, str, Path],
    image_res: Union[int, float],
    mask: Optional[Union[np.ndarray, str, Path]] = None,
    pre_reg_transforms: Optional[dict] = None,
    preprocessing: Optional[ImagePreproParams] = None,
    channel_names: Optional[List[str]] = None,
    channel_colors: Optional[List[str]] = None,
) -> Union[TiffFileRegImage, SitkRegImage, NumpyRegImage, CziRegImage]:
    """
    Convenience function to read in images. Determines the correct reader.

    Parameters
    ----------
    image : str, array-like
        file path to the image to be read or an array like image such as
        a numpy, dask or zarr array
    image_res : float
        spatial resolution of image in units per px (i.e. 0.9 um / px)
    mask: Union[str, Path, np.ndarray]
        path to binary mask (>0 is in) image for registration and/or cropping or a geoJSON with shapes
        that will be processed to a binary mask
    pre_reg_transforms: dict
        Pre-computed transforms to be applied to the image prior to registration
    preprocessing: ImagePreproParams
        preprocessing parameters for the modality for registration. Registration images should be a xy single plane
        so many modalities (multi-channel, RGB) must "create" a single channel.
        Defaults: multi-channel images -> max intensity project image
        RGB -> greyscale then intensity inversion (black background, white foreground)
    channel_names: List[str]
        names for the channels to go into the OME-TIFF
    channel_colors: List[str]
        channels colors for OME-TIFF (not implemented)

    Returns
    -------
    reg_image: RegImage
        A RegImage subclass for the particular image loaded
    """
    if isinstance(image, ARRAYLIKE_CLASSES):
        return NumpyRegImage(
            image,
            image_res,
            mask=mask,
            pre_reg_transforms=pre_reg_transforms,
            preprocessing=preprocessing,
            channel_names=channel_names,
            channel_colors=channel_colors,
            image_filepath=None,
        )

    image_ext = Path(image).suffix.lower()
    if image_ext in TIFFFILE_EXTS:
        reg_image = TiffFileRegImage(
            image,
            image_res,
            mask=mask,
            pre_reg_transforms=pre_reg_transforms,
            preprocessing=preprocessing,
            channel_names=channel_names,
            channel_colors=channel_colors,
        )
    elif image_ext == ".czi":
        reg_image = CziRegImage(
            image,
            image_res,
            mask=mask,
            pre_reg_transforms=pre_reg_transforms,
            preprocessing=preprocessing,
            channel_names=channel_names,
            channel_colors=channel_colors,
        )
    else:
        reg_image = SitkRegImage(
            image,
            image_res,
            mask=mask,
            pre_reg_transforms=pre_reg_transforms,
            preprocessing=preprocessing,
            channel_names=channel_names,
            channel_colors=channel_colors,
        )

    return reg_image



import os
from copy import deepcopy

import numpy as np
import pytest

from wsireg.reg_shapes import RegShapes

HERE = os.path.dirname(__file__)
GEOJSON_FP = os.path.join(HERE, "fixtures/polygons.geojson")


@pytest.mark.usefixtures("complex_transform")
def test_RegShapes_transform(complex_transform):
    rs = RegShapes(GEOJSON_FP)
    shape0 = deepcopy(rs.shape_data[0])
    rs.transform_shapes(complex_transform)
    shape0_tform = rs.transformed_shape_data[0]
    assert np.array_equal(shape0["array"], shape0_tform["array"]) is False


def test_RegShapes_shape_types_numpy():

    points = np.random.randint(1, 100, (2, 20))
    shape_data = []
    for point in points.transpose():
        shape_data.append(
            {
                "array": point,
                "shape_type": 'point',
                "shape_name": 'point',
            }
        )

    ri = RegShapes(shape_data)
    assert len(ri.shape_data_gj[0]["geometry"]["coordinates"]) == 2
    assert ri.shape_data_gj[0]["geometry"]["type"] == "Point"

    shape_data = [
        {
            "array": points,
            "shape_type": 'multipoint',
            "shape_name": 'multipoint',
        }
    ]

    ri = RegShapes(shape_data)
    assert len(ri.shape_data_gj[0]["geometry"]["coordinates"]) == 20
    assert ri.shape_data_gj[0]["geometry"]["type"] == "MultiPoint"

    shape_data = [
        {
            "array": points,
            "shape_type": 'linestring',
            "shape_name": 'linestring',
        }
    ]
    ri = RegShapes(shape_data)

    assert len(ri.shape_data_gj[0]["geometry"]["coordinates"]) == 20
    assert ri.shape_data_gj[0]["geometry"]["type"] == "LineString"


def test_RegShapes_drawmask():
    triangles = [
        np.array([[11, 13], [111, 113], [22, 246]]),
        np.array([[11, 13], [111, 113], [22, 246]]) * 2,
    ]
    rs = RegShapes(triangles)

    bin_mask = rs.draw_mask((512, 512), labels=False)
    assert np.sum(bin_mask) > 0

    lab_mask = rs.draw_mask((512, 512), labels=True)

    # 0 = bg
    assert len(np.unique(lab_mask)) == 3



from pathlib import Path
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
import SimpleITK as sitk
from tifffile import TiffWriter

from wsireg.reg_images.reg_image import RegImage
from wsireg.reg_transforms.reg_transform_seq import RegTransformSeq
from wsireg.utils.im_utils import (
    format_channel_names,
    get_pyramid_info,
    prepare_ome_xml_str,
)


class OmeTiffWriter:

    x_size: Optional[int] = None
    y_size: Optional[int] = None
    y_spacing: Optional[Union[int, float]] = None
    x_spacing: Optional[Union[int, float]] = None
    tile_size: int = 512
    pyr_levels: Optional[List[Tuple[int, int]]] = None
    n_pyr_levels: Optional[int] = None
    PhysicalSizeY: Optional[Union[int, float]] = None
    PhysicalSizeX: Optional[Union[int, float]] = None
    subifds: Optional[int] = None
    compression: str = "deflate"

    def __init__(
        self,
        reg_image: RegImage,
        reg_transform_seq: Optional[RegTransformSeq] = None,
    ):
        """
        Class for managing writing images to OME-TIFF.

        Parameters
        ----------
        reg_image: RegImage
            RegImage to be transformed
        reg_transform_seq: RegTransformSeq or None
            Registration transformation sequence from wsireg to transform image

        Attibutes
        ---------
        x_size: int
            Size of the output image after transformation in x
        y_size: int
            Size of the output image after transformation in y
        y_spacing: float
            Pixel spacing in microns after transformation in y
        x_spacing: float
            Pixel spacing in microns after transformation in x
        tile_size: int
            Size of tiles to be written
        pyr_levels: list of tuples of int:
            Size of downsampled images in pyramid
        n_pyr_levels: int
            Number of downsamples in pyramid
        PhysicalSizeY: float
            physical size of image in micron for OME-TIFF in Y
        PhysicalSizeX: float
            physical size of image in micron for OME-TIFF in X
        subifds: int
            Number of sub-resolutions for pyramidal OME-TIFF
        compression: str
            tifffile string to pass to compression argument, defaults to "deflate" for minisblack
            and "jpeg" for RGB type images

        """
        self.reg_image = reg_image
        self.reg_transform_seq = reg_transform_seq

    def _prepare_image_info(
        self,
        image_name: str,
        reg_transform_seq: Optional[RegTransformSeq] = None,
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: Optional[str] = "default",
    ) -> None:
        """Get image info and OME-XML"""

        if reg_transform_seq:
            self.x_size, self.y_size = reg_transform_seq.output_size
            self.x_spacing, self.y_spacing = reg_transform_seq.output_spacing
        else:
            self.y_size, self.x_size = (
                (self.reg_image.shape[0], self.reg_image.shape[1])
                if self.reg_image.is_rgb
                else (self.reg_image.shape[1], self.reg_image.shape[2])
            )
            self.y_spacing, self.x_spacing = None, None

        self.tile_size = tile_size
        # protect against too large tile size
        while (
            self.y_size / self.tile_size <= 1
            or self.x_size / self.tile_size <= 1
        ):
            self.tile_size = self.tile_size // 2

        self.pyr_levels, _ = get_pyramid_info(
            self.y_size, self.x_size, self.reg_image.n_ch, self.tile_size
        )
        self.n_pyr_levels = len(self.pyr_levels)

        if reg_transform_seq:
            self.PhysicalSizeY = self.y_spacing
            self.PhysicalSizeX = self.x_spacing
        else:
            self.PhysicalSizeY = self.reg_image.image_res
            self.PhysicalSizeX = self.reg_image.image_res

        channel_names = format_channel_names(
            self.reg_image.channel_names, self.reg_image.n_ch
        )

        self.omexml = prepare_ome_xml_str(
            self.y_size,
            self.x_size,
            self.reg_image.n_ch,
            self.reg_image.im_dtype,
            self.reg_image.is_rgb,
            PhysicalSizeX=self.PhysicalSizeX,
            PhysicalSizeY=self.PhysicalSizeY,
            PhysicalSizeXUnit="µm",
            PhysicalSizeYUnit="µm",
            Name=image_name,
            Channel=None if self.reg_image.is_rgb else {"Name": channel_names},
        )

        self.subifds = self.n_pyr_levels - 1 if write_pyramid is True else None

        if compression == "default":
            print("using default compression")
            self.compression = "jpeg" if self.reg_image.is_rgb else "deflate"
        else:
            self.compression = compression

    def write_image_by_plane(
        self,
        image_name: str,
        output_dir: Union[Path, str] = "",
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: Optional[str] = "default",
    ) -> str:
        """
        Write OME-TIFF image plane-by-plane to disk. WsiReg compatible RegImages all
        have methods to read an image channel-by-channel, thus each channel is read, transformed, and written to
        reduce memory during write.
        RGB images may run large memory footprints as they are interleaved before write, for RGB images,
        using the `OmeTiledTiffWriter` is recommended.

        Parameters
        ----------
        image_name: str
            Name to be written WITHOUT extension
            for example if image_name = "cool_image" the file
            would be "cool_image.ome.tiff"
        output_dir: Path or str
            Directory where the image will be saved
        write_pyramid: bool
            Whether to write the OME-TIFF with sub-resolutions or not
        tile_size: int
            What size to write OME-TIFF tiles to disk
        compression: str
            tifffile string to pass to compression argument, defaults to "deflate" for minisblack
            and "jpeg" for RGB type images

        Returns
        -------
        output_file_name: str
            File path to the written OME-TIFF

        """

        output_file_name = str(Path(output_dir) / f"{image_name}.ome.tiff")
        self._prepare_image_info(
            image_name,
            reg_transform_seq=self.reg_transform_seq,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
        )

        rgb_im_data = []

        print(f"saving to {output_file_name}")
        with TiffWriter(output_file_name, bigtiff=True) as tif:
            if self.reg_image.reader == "sitk":
                self.reg_image._read_full_image()

            for channel_idx in range(self.reg_image.n_ch):
                print(f"transforming : {channel_idx}")
                image = self.reg_image.read_single_channel(channel_idx)
                image = np.squeeze(image)
                image = sitk.GetImageFromArray(image)
                image.SetSpacing(
                    (self.reg_image.image_res, self.reg_image.image_res)
                )

                if self.reg_transform_seq:
                    image = self.reg_transform_seq.resampler.Execute(image)
                    # image = transform_plane(
                    #     image, final_transform, composite_transform
                    # )
                    print(f"transformed : {channel_idx}")

                if self.reg_image.is_rgb:
                    rgb_im_data.append(image)
                else:
                    print("saving")
                    if isinstance(image, sitk.Image):
                        image = sitk.GetArrayFromImage(image)

                    options = dict(
                        tile=(self.tile_size, self.tile_size),
                        compression=self.compression,
                        photometric="rgb"
                        if self.reg_image.is_rgb
                        else "minisblack",
                        metadata=None,
                    )
                    # write OME-XML to the ImageDescription tag of the first page
                    description = self.omexml if channel_idx == 0 else None
                    # write channel data
                    print(
                        f" writing channel {channel_idx} - shape: {image.shape}"
                    )
                    tif.write(
                        image,
                        subifds=self.subifds,
                        description=description,
                        **options,
                    )

                    if write_pyramid:
                        for pyr_idx in range(1, self.n_pyr_levels):
                            resize_shape = (
                                self.pyr_levels[pyr_idx][0],
                                self.pyr_levels[pyr_idx][1],
                            )
                            image = cv2.resize(
                                image,
                                resize_shape,
                                cv2.INTER_LINEAR,
                            )
                            print(
                                f"pyramid index {pyr_idx} : channel {channel_idx} shape: {image.shape}"
                            )

                            tif.write(image, **options, subfiletype=1)

            if self.reg_image.is_rgb:
                rgb_im_data = sitk.Compose(rgb_im_data)
                rgb_im_data = sitk.GetArrayFromImage(rgb_im_data)

                options = dict(
                    tile=(self.tile_size, self.tile_size),
                    compression=self.compression,
                    photometric="rgb",
                    metadata=None,
                )
                # write OME-XML to the ImageDescription tag of the first page
                description = self.omexml

                # write channel data
                tif.write(
                    rgb_im_data,
                    subifds=self.subifds,
                    description=description,
                    **options,
                )

                print(f"RGB shape: {rgb_im_data.shape}")
                if write_pyramid:
                    for pyr_idx in range(1, self.n_pyr_levels):
                        resize_shape = (
                            self.pyr_levels[pyr_idx][0],
                            self.pyr_levels[pyr_idx][1],
                        )
                        rgb_im_data = cv2.resize(
                            rgb_im_data,
                            resize_shape,
                            cv2.INTER_LINEAR,
                        )
                        tif.write(rgb_im_data, **options, subfiletype=1)
        return output_file_name



import os
from pathlib import Path


HERE = os.path.dirname(__file__)
FIXTURES_DIR = os.path.join(HERE, "fixtures")
PRIVATE_DIR = os.path.join(HERE, "private_data")

config1_fp = str(Path(FIXTURES_DIR) / "test-config1-cmd-line.yaml")


def test_wsireg2d_entrypoint():
    exit_status = os.system('wsireg2d --help')
    assert exit_status == 0


def test_wsireg2d_run():
    exit_status = os.system(f'wsireg2d "{str(config1_fp)}" --testing')
    assert exit_status == 0
